#!/usr/bin/python3
# -*- coding: UTF-8 -*-

# 28.08.2013

# Kompassmodul LSM303 auslesen, Heading ermitteln und grafisch anzeigen
# Es gibt die Option:
# - die Acceleration-Werte bei der Bestimmung der Richtung berücksichtigen
#   seitliche Verschiebungen/Beschleunigungen verfälschen den Messwert dann
#   deutlich.
# - die Acceleration-Wert konstant als (0,0,-1) zu setzen,
#   also von einer sauberen, horizontalen Ausrichtung auszugehen.
#   Neigung des Sensors verfälscht in diesem Fall den Messwert .....

# Die Berechnungen sind aus C-Programmen übernommen, die ich hier gefunden habe:
# File: LSM303DLH-orangutan-example-code.zip
#
# Copyright (c) 2011 Pololu Corporation.  For more information, see
# http://www.pololu.com/
# http://forum.pololu.com/


import time, math, _thread
from tkinter import *
from atmega_p import ATMega_p
from ser_port import open_ser_port
from global_h import *
from lsm303_h import *
from compass import Compass


class LSM303(object):
    """ Sensor module LSM303DLM
        3-axis accelerometer and 3-axis magnetometer """
    def __init__(self):
        self.__twi_adr_a = LSM303_TWI_ADR_A
        self.__twi_adr_m = LSM303_TWI_ADR_M
        self.m_min = [0,0,0]
        self.m_max = [0,0,0]
        self.m_null = [0,0,0]
        self._calc_acceleration = True

    # accelerometer ----
    def write_register_a(self, reg, data):
        """ schreibt genau ein Byte (accelerometer) """
        uc.twiwrite(self.__twi_adr_a, regadr = reg, data = data)


    def read_register_a(self, reg, cnt = 1):
        """ liest Bytes vom accelerometer aus """
        return uc.twiread(self.__twi_adr_a, regadr = reg, cnt = cnt)


    def read_data_a(self):
        """ Liest die Beschleunigungswerte aus

            Die Reihenfolge high-Byte, low-Byte kann in der
            Konfiguration eingestellt werden
            im Adressregister LSM_OUT_X_L_A muss das höchstwertige
            Bit gesetzt werden, damit ein Autoincrement ausgeführt wird !"""

        t = [0,0,0]
        a = self.read_register_a(reg = LSM_OUT_X_L_A + 0x80, cnt = 6)

        for i in range(3):
            t[i] = a[i * 2] << 8
            t[i] += a[(i * 2) + 1]

            if t[i] > 32768:
                t[i] -= 65536
        return t


    # magnetometer ------------
    def write_register_m(self, reg, data):
        """ schreibt genau ein Byte in ein Register (Magnetic) """
        uc.twiwrite(self.__twi_adr_m, regadr = reg, data = data)


    def read_data_m(self):
        """ Liest den MagnetSensor aus """

        t = uc.twiread(self.__twi_adr_m, regadr = LSM_OUT_X_H_M, cnt = 6)
        acc = [0,0,0]

        for i in range(3):
            acc[i]  = (t[i*2] << 8) + t[(i * 2) +1]
            # von unsigned zu integer
            if acc[i] > 32768:
                acc[i] -= 65536

        return acc


    def init_register(self):
        """ Die Register des Sensors konfigurieren """
        # accelerometer
        self.write_register_a(LSM_CTRL_REG1_A, 0x27)
        self.write_register_a(LSM_CTRL_REG4_A, 0xC0)

        # magnetic
        self.write_register_m(LSM_CRA_REG_M, 0x14)
        self.write_register_m(LSM_CRB_REG_M, 0x20)
        self.write_register_m(LSM_MR_REG_M, 0x00)


##    def print_registers(self):
##        """ Gibt zum Debuggen die Werte der relevanten Register aus """
##        for i in range(0x1D, 0x3A, 1):
##            r = self.read_register(i)
##            print(hex(i), hex(r[0]))


    def calibrate_m(self):
        """ Min- und Max-Werte der Magnetfeldmessung ermitteln
            Die Werte werden anschließend als Konstanten manuell
            in den Programmcode übertragen -> read_calibration()
            Der Sensor muss mit X,Y und Z-Achse in Richtung des einfallenden
            Magnetfeldes gedreht werden (und natürlich in die Gegenrichtung),
            damit die Extremwerte möglichst gut erfasst werden """
        t = 500         # 500 Messungen durchführen
        while t:
            t -= 1
            m = self.read_data_m()

            # X-Komponente
            if m[0] > self.m_max[0]:
                self.m_max[0] = m[0]
            if m[0] < self.m_min[0]:
                self.m_min[0] = m[0]

            # Y-Komponente
            if m[2] > self.m_max[1]:
                self.m_max[1] = m[2]
            if m[2] < self.m_min[1]:
                self.m_min[1] = m[2]

            # Z-Komponente
            if m[1] > self.m_max[2]:
                self.m_max[2] = m[1]
            if m[1] < self.m_min[2]:
                self.m_min[2] = m[1]

            print(self.m_min, self.m_max)
            time.sleep(.05)

        self.m_null[0] = (self.m_min[0] + self.m_max[0]) // 2
        self.m_null[1] = (self.m_min[1] + self.m_max[1]) // 2
        self.m_null[2] = (self.m_min[2] + self.m_max[2]) // 2

        print("     Min  Null   Max")
        print("X: {:5} {:5} {:5}". format(self.m_min[0], self.m_null[0], self.m_max[0]))
        print("Y: {:5} {:5} {:5}". format(self.m_min[1], self.m_null[1], self.m_max[1]))
        print("Z: {:5} {:5} {:5}". format(self.m_min[2], self.m_null[2], self.m_max[2]))
        print()
        print("Min- und Max-Werte übernehmen in read_calibration() !")
        exit()


    def read_calibration(self):
        """ Die Extremwerte werden mit der Funktion calibrate() ermittelt
            und anschließend hier eingetragen.
            Beim normalen Programmstart werden die Werte hier eingelesen
            Diese Werte sind typisch für einen Sensor und seine Umgebung """
        self.m_min = [-483, -585, -581]
        self.m_max = [421, 342, 257]


    def read_avg_a(self, cnt = 8):
        """ Mittelwert Accelerometer aus mehreren Messungen bilden """
        avg = [0,0,0]
        i = cnt
        while i > 0:
            i -= 1
            a = self.read_data_a()
            avg[0] += a[0]
            avg[1] += a[1]
            avg[2] += a[2]
            time.sleep(.05)
        avg[0] = avg[0] // cnt
        avg[1] = avg[1] // cnt
        avg[2] = avg[2] // cnt

        return avg

    def read_avg_m(self, cnt = 8):
        """ Mittelwert Magnetometer aus mehreren Messungen bilden """
        avg = [0,0,0]
        i = cnt
        while i > 0:
            i -= 1
            a = self.read_data_m()
            avg[0] += a[0]
            avg[1] += a[2]
            avg[2] += a[1]
            time.sleep(.05)
        avg[0] = avg[0] // cnt
        avg[1] = avg[1] // cnt
        avg[2] = avg[2] // cnt

        return avg

    def set_calc_acceleration(self, v):
        """ Setzt das Flag self._calc_acceleration auf True / False """
        if v:
            self._calc_acceleration = True
        else:
            self._calc_acceleration = False


    def heading(self):
        """ Berechnet die Ausrichtung des Sensors und liefert den Winkel gegen
            Nord zurück
            - mit bzw. ohne Berücksichtigung der Neigung/Accelerometer-Werte """

        if self._calc_acceleration is True:
            a = lsm.read_avg_a(cnt = 2)     # Neigung ermitteln
        else:
            a = [0,0,-1]                    # von horizontaler Lage ausgehen
        m = lsm.read_avg_m(cnt = 8)

        m[0] = ((m[0] - self.m_min[0]) / (self.m_max[0] - self.m_min[0])) * 2 - 1.0
        m[1] = ((m[1] - self.m_min[1]) / (self.m_max[1] - self.m_min[1])) * 2 - 1.0
        m[2] = ((m[2] - self.m_min[2]) / (self.m_max[2] - self.m_min[2])) * 2 - 1.0

        vE = vector_cross(a, m)
        vE = vector_normalize(vE)

        vN = vector_cross(a, vE)
        vN = vector_normalize(vN)

        #p = (0, -1, 0)
        h = int(math.atan2(vector_dot(vE, (0, -1, 0)), vector_dot(vN, (0,-1, 0))) * 180 / math.pi)
        if h < 0:
            h += 360
        return h


if __name__ == "__main__":

    def vector_dot(v, w):
        return (v[0] * w[0] + v[1] * w[1] + v[2] * w[2])

    def vector_normalize(v):
        a = v[0]**2 + v[1]**2 + v[2]**2
        a = a**0.5
        v[0] /= a
        v[1] /= a
        v[2] /= a
        return v

    def vector_cross(v, w):
        a = [0,0,0]
        a[0] = v[1] * w[2] - v[2] * w[1]
        a[1] = v[2] * w[0] - v[0] * w[2]
        a[2] = v[0] * w[1] - v[1] * w[0]

        return a

    def animate():
        """ Hintergrundthread, der regelmäßig die Anzeige aktualisiert """
        try:
            while 1:
                time.sleep(.1)
                cp.rotate(lsm.heading())
        except:
            exit()

    def checkbuttonchanged():
        """ wird aufgerufen, wenn der Checkbutton geändert wird """
        #print(lsmstate.get())
        lsm.set_calc_acceleration(lsmstate.get())

    uc = ATMega_p(open_ser_port())

    lsm = LSM303()
    lsm.init_register()
    #lsm.calibrate_m()      # wird nur benutzt, um die Kalibrierungswerte zu gewinnen
                            # die dann in die Funktion read_calibration als
                            # Konstanten eingetragen werden
    lsm.read_calibration()

    top = Tk()
    top.title("LSM303")
    top.resizable(width = False, height = False) # Größe nicht veränderbar

    cp = Compass(top)

    lsmstate = IntVar()
    lsmstate.set(1)

    btn = Checkbutton(top, text = "Die Sensorneigung wird rechnerisch eliminiert",
            variable = lsmstate, font = ("Helvetica", "10"),command = checkbuttonchanged)
    btn.grid(column = 0, row = 1, padx = 10, pady = 10, sticky = "NESW")

    _thread.start_new_thread(animate,())

    top.mainloop()
